
%  Copyright (c) 2013
%      Noa Fish <noafish@post.tau.ac.il>

%%% cast a ray between a source and target points and return the closest
%%% hit point

%%% input:
% points: 3D point set
% normals: 3D normals corresponding to points
% splatRadiusesSquare: estimation of point splat radii (helps determine which points are intersected by a ray)
% faceFronters: the points that are in front of the source point
% src: source point
% tgt: target point
% gr: KNN graph
% outliers: outlier points

%%% output:
% hit_point: index of the closest hit point


function hit_point = cast_ray(points, normals, splatRadiusesSquare, faceFronters, src, tgt, gr, outliers)

almostZero = 1e-10;
hit_point = -1;

srcp = points(src,:);
srcn = normals(src,:);
tgtp = points(tgt,:);
tgtn = normals(tgt,:);

vec = tgtp - srcp;
vec = vec / norm(vec);

dists = bsxfun(@minus, points, srcp);
dists = sum(dists .^ 2, 2);
dists = sqrt(dists);
closer = find(dists <= dists(tgt));
orig_points = points;
points = points(closer,:);
normals = normals(closer,:);
splatRadiusesSquare = splatRadiusesSquare(closer);
orig_src_idx = find(closer == src);
orig_tgt_idx = find(closer == tgt);

rayVector = vec;

% find splats that are intersected by the ray vector
% this is found using line to plane distance formula
linePlaneFormulaNominator   = sum(bsxfun(@minus, points, srcp) .* normals, 2);
linePlaneFormulaDenominator = sum(bsxfun(@times, rayVector, normals), 2);

valid = valid_point(tgt, orig_tgt_idx, faceFronters, linePlaneFormulaDenominator, almostZero);
if(~valid)
    return;
end

% allow only points that are in front of the face
rayIntersectors = abs(linePlaneFormulaDenominator) > almostZero; % true on splats that the ray intersects (and not contained within or parallel to)
rayIntersectors(orig_src_idx) = false; % never consider the source as a possible hit

numIntersectors = sum(rayIntersectors);
if(numIntersectors == 0)
    return;
end


% now we care only about intersecting-splats
rayIntersectorsIndex = find(rayIntersectors);
intersectPoints = points(rayIntersectors,:);
intersectSplatRadiusesSquare = splatRadiusesSquare(rayIntersectors);

intersectLinePlaneFormulaNominator = linePlaneFormulaNominator(rayIntersectors,:);
intersectLinePlaneFormulaDenominator = linePlaneFormulaDenominator(rayIntersectors,:);

% calculate points of intersections to each point's plane
lengthOfRays = intersectLinePlaneFormulaNominator ./ intersectLinePlaneFormulaDenominator;
pointsOfIntersection = bsxfun(@plus, srcp, bsxfun(@times, lengthOfRays, rayVector));

% take only points where the intersection is close to the point
distancesToIntersectionSquare = sum((intersectPoints - pointsOfIntersection).^2,2);                        
intersectionsInSplatRadius = distancesToIntersectionSquare <= intersectSplatRadiusesSquare;
lengthOfRays(~intersectionsInSplatRadius) = inf;

if(isempty(find(gr{src} == tgt)))
    distancesFromSourceSquare = sum(bsxfun(@minus, intersectPoints, srcp) .^ 2, 2);
    distantFromSource = distancesFromSourceSquare > 4 * intersectSplatRadiusesSquare;
    lengthOfRays(~distantFromSource) = inf;
end

noninf = find(lengthOfRays ~= Inf);
noninfdist = lengthOfRays(noninf);
noninfdist = abs(noninfdist);
hits = [noninf noninfdist];
hits = sortrows(hits,2);

for t=1:size(hits,1)
    ind = rayIntersectorsIndex(hits(t,1));
    ind = closer(ind);
    if(ind == tgt)
        hit_point = tgt;
        break;
    end

    % if hit point is a neighbor of tgt, consider it as a good hit
    is_neigh = find(gr{tgt} == ind, 1);
    if(~isempty(is_neigh))
        
        hit_point = tgt;
        break;
    end

    % make sure that the hit point is on the right side of the ray and not
    % on the opposite direction
    hit_vec = orig_points(ind,:) - srcp;
    hit_vec = hit_vec / norm(hit_vec);
    ang = vec_angle(vec, hit_vec);
    if(abs(ang) < 0.05 && isempty(find(outliers == ind)))
        hit_point = ind;
        break;
    end
end


if(hit_point ~= -1)
    is_neigh = find(gr{src} == hit_point, 1);
    ints = intersect(gr{src}, gr{hit_point});
    
    % if the hit point is not tgt and either the hit point is a nearest
    % neighbor of src or it has mutual nearest neighbors with src, return
    % -1 as the hit point so that the geodesic path between src and tgt can
    % be checked
    if(hit_point ~= tgt && (~isempty(is_neigh) || ~isempty(ints)))
        hit_point = -1;
    end
end


end



function valid = valid_point(tgt, orig_tgt_idx, faceFronters, linePlaneFormulaDenominator, almostZero)

valid = 1;

if(~faceFronters(tgt) || linePlaneFormulaDenominator(orig_tgt_idx) <= almostZero)
    valid = 0;
end

end





